<?php

namespace Drupal\commerce_payone\Plugin\Commerce\PaymentGateway;

use Drupal\address\AddressInterface;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_payment\Exception\InvalidRequestException;
use Drupal\commerce_payment\PaymentMethodTypeManager;
use Drupal\commerce_payment\PaymentTypeManager;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\OffsitePaymentGatewayBase;
use Drupal\commerce_payone\PayoneApiServiceInterface;
use Drupal\commerce_price\Calculator;
use Drupal\commerce_price\Price;
use Drupal\commerce_product\Entity\ProductVariationInterface;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelTrait;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;
use Symfony\Component\HttpFoundation\Request;

/**
 * Abstract base class for any Payone offsite payment gateway.
 */
abstract class PayoneOffsitePaymentGatewayBase extends OffsitePaymentGatewayBase {

  use LoggerChannelTrait;
  use PayonePaymentGatewayConfigurationTrait;

  /**
   * The gateway used for making API calls.
   *
   * @var \Drupal\commerce_payone\PayoneApiServiceInterface
   */
  protected $api;

  /**
   * The event dispatcher.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected $eventDispatcher;

  /**
   * The payment storage.
   *
   * @var \Drupal\commerce_payment\PaymentStorageInterface
   */
  protected $paymentStorage;

  /**
   * Constructs a new PayoneOffsitePaymentGatewayBase object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\commerce_payment\PaymentTypeManager $payment_type_manager
   *   The payment type manager.
   * @param \Drupal\commerce_payment\PaymentMethodTypeManager $payment_method_type_manager
   *   The payment method type manager.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time.
   * @param \Drupal\commerce_payone\PayoneApiServiceInterface $api_service
   *   The Payone API service.
   * @param \Symfony\Component\EventDispatcher\EventDispatcherInterface $event_dispatcher
   *   The event dispatcher.
   */
  public function __construct(
    array $configuration,
    string $plugin_id,
    $plugin_definition,
    EntityTypeManagerInterface $entity_type_manager,
    PaymentTypeManager $payment_type_manager,
    PaymentMethodTypeManager $payment_method_type_manager,
    TimeInterface $time,
    PayoneApiServiceInterface $api_service,
    EventDispatcherInterface $event_dispatcher
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $entity_type_manager, $payment_type_manager, $payment_method_type_manager, $time);

    $this->api = $api_service;
    $this->eventDispatcher = $event_dispatcher;
    $this->paymentStorage = $entity_type_manager->getStorage('commerce_payment');
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('plugin.manager.commerce_payment_type'),
      $container->get('plugin.manager.commerce_payment_method_type'),
      $container->get('datetime.time'),
      $container->get('commerce_payone.payment_api'),
      $container->get('event_dispatcher')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return $this->defaultPayonePaymentGatewayConfiguration()
      + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);
    return $this->buildPayonePaymentGatewayConfigurationForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);
    $this->submitPayonePaymentGatewayConfigurationForm($form, $form_state, $this->configuration);
  }

  /**
   * {@inheritdoc}
   */
  public function onCancel(OrderInterface $order, Request $request) {
    parent::onCancel($order, $request);

    $payment_gateway_data = $order->getData($this->getOrderDataKey());
    if (empty($payment_gateway_data['transaction_id'])) {
      throw new InvalidRequestException('Transaction ID missing for this transaction.');
    }

    $transaction_id = $payment_gateway_data['transaction_id'];
    $payment = $this->paymentStorage->loadByRemoteId($transaction_id);
    if (empty($payment)) {
      throw new InvalidRequestException('No transaction found for transaction ID @transaction_id.', ['@transaction_id' => $transaction_id]);
    }

    $payment->state = 'authorization_voided';
    $payment->save();
  }

  /**
   * Returns an array populated with basic authorization request data.
   *
   * The array already includes the call to getServerApiStandardParameters().
   *
   * Payment method specific additional data like wallettype must be populated
   * in the payment gateway itself.
   *
   * @param \Drupal\commerce_payment\Entity\PaymentInterface $payment
   *   The payment entity.
   * @param \Drupal\address\AddressInterface $billing_address
   *   The billing address.
   *
   * @return array
   *   An array populated with basic data needed for any payment type.
   */
  protected function getBasicServerApiAuthorizationParameters(PaymentInterface $payment, AddressInterface $billing_address) {
    $order = $payment->getOrder();
    $owner = $order->getCustomer();
    if ($owner && $owner->isAuthenticated()) {
      $customer_id = $this->getRemoteCustomerId($owner);
      $customer_email = $owner->getEmail();
    }
    else {
      $customer_email = $order->getEmail();
    }

    $parameters = [
      'clearingtype' => $this->getClearingType(),
      'aid' => $this->configuration['sub_account_id'],
      // Reference must be unique.
      'reference' => $payment->getOrderId() . '_' . $this->time->getCurrentTime(),
      'amount' => $this->toMinorUnits($payment->getAmount()),
      'currency' => $payment->getAmount()->getCurrencyCode(),
      'email' => $customer_email,
      'firstname' => $billing_address->getGivenName(),
      'lastname' => $billing_address->getFamilyName(),
      'company' => $billing_address->getOrganization(),
      'street' => $billing_address->getAddressLine1(),
      'addressaddition' => $billing_address->getAddressLine2(),
      'zip' => $billing_address->getPostalCode(),
      'city' => $billing_address->getLocality(),
      'country' => $billing_address->getCountryCode(),
      'ip' => $order->getIpAddress(),
    ];

    if (!empty($customer_id)) {
      $parameters['userid'] = $customer_id;
    }

    if ($owner && $owner->isAuthenticated()) {
      $parameters['customerid'] = $owner->id();
    }

    if ($shipping_profile = $this->getShippingProfile($order)) {
      if ($shipping_address = $shipping_profile->get('address')->first()) {
        $parameters['shipping_firstname'] = $shipping_address->getGivenName();
        $parameters['shipping_lastname'] = $shipping_address->getFamilyName();
        $parameters['shipping_company'] = $shipping_address->getOrganization();
        $parameters['shipping_street'] = $shipping_address->getAddressLine1();
        $parameters['shipping_addressaddition'] = $shipping_address->getAddressLine2();
        $parameters['shipping_zip'] = $shipping_address->getPostalCode();
        $parameters['shipping_city'] = $shipping_address->getLocality();
        $parameters['shipping_country'] = $shipping_address->getCountryCode();
      }
    }

    /** @var \Drupal\commerce_price\Price[] $promotion_amounts_by_vat_rate */
    $promotion_amounts_by_vat_rate = [];

    $item_id = 1;
    foreach ($order->getItems() as $item) {
      $purchased_entity = $item->getPurchasedEntity();
      $parameters[sprintf('it[%s]', $item_id)] = 'goods';
      $sku = $item->label();
      if ($purchased_entity instanceof ProductVariationInterface) {
        $sku = $purchased_entity->getSku();
      }
      elseif ($purchased_entity instanceof FieldableEntityInterface && $purchased_entity->hasField('sku') && !$purchased_entity->get('sku')->isEmpty()) {
        $sku = $purchased_entity->sku->value;
      }
      $parameters[sprintf('id[%s]', $item_id)] = $sku;
      $parameters[sprintf('pr[%s]', $item_id)] = $this->toMinorUnits($item->getAdjustedUnitPrice(['tax']));
      $parameters[sprintf('no[%s]', $item_id)] = Calculator::trim($item->getQuantity());
      $parameters[sprintf('de[%s]', $item_id)] = $item->label();
      $tax_adjustments = $item->getAdjustments(['tax']);
      $vat_rate = !empty($tax_adjustments) ? reset($tax_adjustments)->getPercentage() * 100 : 0;
      $parameters[sprintf('va[%s]', $item_id)] = $vat_rate;
      if ($promotion_adjustments = $item->getAdjustments(['promotion'])) {
        if (!isset($promotion_amounts_by_vat_rate[$vat_rate])) {
          $promotion_amounts_by_vat_rate[$vat_rate] = new Price('0', $item->getTotalPrice()->getCurrencyCode());
        }
        foreach ($promotion_adjustments as $promotion_adjustment) {
          $promotion_amounts_by_vat_rate[$vat_rate] = $promotion_amounts_by_vat_rate[$vat_rate]->add($promotion_adjustment->getAmount());
        }
      }
      $item_id++;
    }

    if ($order->hasField('shipments') && !$order->get('shipments')->isEmpty()) {
      /** @var \Drupal\commerce_shipping\Entity\ShipmentInterface[] $shipments */
      $shipments = $order->get('shipments')->referencedEntities();
      foreach ($shipments as $shipment) {
        $parameters[sprintf('it[%s]', $item_id)] = 'shipment';
        $label = $this->t('Shipping');
        $parameters[sprintf('id[%s]', $item_id)] = 'shipping';
        // Always take the adjusted amount for shipping, as we don't get the
        // promotions via $shipment->getAdjustments() reliably.
        $parameters[sprintf('pr[%s]', $item_id)] = $this->toMinorUnits($shipment->getAdjustedAmount());
        $parameters[sprintf('no[%s]', $item_id)] = 1;
        $parameters[sprintf('de[%s]', $item_id)] = $label;
        $tax_adjustments = $shipment->getAdjustments(['tax']);
        $vat_rate = !empty($tax_adjustments) ? reset($tax_adjustments)->getPercentage() * 100 : 0;
        $parameters[sprintf('va[%s]', $item_id)] = $vat_rate;
        $item_id++;
      }
    }

    foreach ($promotion_amounts_by_vat_rate as $vat_rate => $promotion_amount) {
      $parameters[sprintf('it[%s]', $item_id)] = 'voucher';
      $label = $this->t('Discount');
      $parameters[sprintf('id[%s]', $item_id)] = 'discount';
      $parameters[sprintf('pr[%s]', $item_id)] = $this->toMinorUnits($promotion_amount);
      $parameters[sprintf('no[%s]', $item_id)] = 1;
      $parameters[sprintf('de[%s]', $item_id)] = $label;
      $parameters[sprintf('va[%s]', $item_id)] = $vat_rate;
    }

    return array_merge($this->api->getServerApiStandardParameters($this->configuration, 'authorization'), $parameters);
  }

  /**
   * Get payment redirect key.
   *
   * @param int $order_id
   *   The order id.
   *
   * @return string
   *   The redirect key.
   */
  protected function getPaymentRedirectKey(int $order_id) {
    $data = $this->api->getServerApiStandardParameters($this->configuration, 'authorization');
    $data['order_id'] = $order_id;
    return $this->api->generateHash($data, $this->configuration['key'], $this->configuration['hash_method']);
  }

  /**
   * Build the return URL including hashed 'key' parameter.
   *
   * @param string $url
   *   The basic url.
   * @param int $order_id
   *   The order id.
   *
   * @return \Drupal\Core\Url
   *   The url.
   */
  protected function buildReturnUrl(string $url, int $order_id) {
    return Url::fromUri($url, [
      'absolute' => TRUE,
      'query' => ['key' => $this->getPaymentRedirectKey($order_id)],
    ]);
  }

  /**
   * Gets the shipping profile from the given order, if available.
   *
   * @param \Drupal\commerce_order\Entity\OrderInterface $order
   *   The order entity.
   *
   * @return \Drupal\profile\Entity\ProfileInterface|null
   *   The shipping profile, if available - NULL otherwise.
   */
  protected function getShippingProfile(OrderInterface $order) {
    if ($order->hasField('shipments')) {
      /** @var \Drupal\commerce_shipping\Entity\ShipmentInterface $shipment */
      foreach ($order->shipments->referencedEntities() as $shipment) {
        return $shipment->getShippingProfile();
      }
    }

    return NULL;
  }

  /**
   * Get the clearing type of this plugin.
   *
   * @return string
   *   The clearing type of this plugin to be used for API requests.
   */
  abstract protected function getClearingType();

  /**
   * Gets the key to be used to store Gateway specific info in the order data.
   *
   * @return string
   *   The key to be used to store Gateway specific info in the order data.
   */
  abstract public function getOrderDataKey();

}
