<?php

namespace Drupal\commerce_payone\Plugin\Commerce\PaymentGateway;

use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_payment\Exception\HardDeclineException;
use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payone\Event\CommercePayoneEvents;
use Drupal\commerce_payone\Event\PayoneRequestEvent;
use Symfony\Component\HttpFoundation\Request;

/**
 * Provides the Off-site Redirect payment gateway.
 *
 * @CommercePaymentGateway(
 *   id = "payone_sofort",
 *   label = "Payone Sofortüberweisung (Off-site)",
 *   display_label = "Payone Sofortüberweisung",
 *   forms = {
 *     "offsite-payment" = "Drupal\commerce_payone\PluginForm\SofortBanking\SofortBankingOffsiteForm",
 *   },
 *   payment_method_types = {"commerce_payone_sofort"},
 * )
 */
class PayoneSofort extends PayoneOffsitePaymentGatewayBase implements PayoneSofortInterface {

  /**
   * {@inheritdoc}
   */
  protected function getClearingType() {
    return 'sb';
  }

  /**
   * {@inheritdoc}
   */
  public function getOrderDataKey() {
    return 'payone_sofort';
  }

  /**
   * {@inheritdoc}
   */
  public function initializeSofortApi(PaymentInterface $payment, array $form) {
    /** @var \Drupal\address\AddressInterface $billing_address */
    if ($payment->getOrder()->getBillingProfile()) {
      $billing_address = $payment->getOrder()->getBillingProfile()->get('address')->first();
    }
    if (empty($billing_address)) {
      throw new HardDeclineException('No billing address defined for the payment.');
    }

    $request = $this->getBasicServerApiAuthorizationParameters($payment, $billing_address);
    $request['onlinebanktransfertype'] = 'PNT';
    $request['bankcountry'] = $billing_address->getCountryCode();
    $request['successurl'] = $this->buildReturnUrl($form['#return_url'], $payment->getOrderId())->toString();
    $request['backurl'] = $form['#cancel_url'];
    $request['errorurl'] = $form['#cancel_url'];

    // Allow modules to adjust the request.
    $event = new PayoneRequestEvent($payment, $request);
    $this->eventDispatcher->dispatch(CommercePayoneEvents::PayoneInitializeRequestEvent, $event);
    $request = $event->getRequest();

    return $this->api->processHttpPost($request, FALSE);
  }

  /**
   * {@inheritdoc}
   */
  public function onReturn(OrderInterface $order, Request $request) {
    $payment_gateway_data = $order->getData($this->getOrderDataKey());
    $transaction_id = $payment_gateway_data['transaction_id'];

    if (empty($transaction_id)) {
      throw new PaymentGatewayException('Transaction ID missing for transaction.');
    }

    if (empty($request->query->get('key')) || $request->query->get('key') !== $this->getPaymentRedirectKey($order->id())) {
      throw new PaymentGatewayException('Payment could not be authorized.');
    }

    $payment = $this->paymentStorage->loadByRemoteId($transaction_id);
    if (empty($payment)) {
      $this->getLogger('commerce_payone')->warning('No payment found for remote ID @transaction_id. A new payment entity will be created instead.', ['@transaction_id' => $transaction_id]);

      $payment = $this->paymentStorage->create([
        'amount' => $order->getTotalPrice(),
        'payment_gateway' => $this->parentEntity->id(),
        'order_id' => $order->id(),
        'remote_id' => $transaction_id,
      ]);
    }
    $payment->setState('completed');
    $payment->setRemoteState('APPROVED');
    $payment->save();
  }

}
